#points to commands that help request info/actions using the gateway
from .guild.request import GuildRequest
from .dms.request import DmRequest
from .media.request import MediaRequest

class Request(object):
	def __init__(self, gatewayobject):
		self.gatewayobject = gatewayobject

	def lazyGuild(self, guild_id, channel_ranges, typing=None, threads=None, activities=None, members=None):
		GuildRequest(self.gatewayobject).lazyGuild(guild_id, channel_ranges, typing, threads, activities, members)

	def searchGuildMembers(self, guild_ids, query, limit=10, presences=True):
		GuildRequest(self.gatewayobject).searchGuildMembers(guild_ids, query, limit, presences)

	def DMchannel(self, channel_id):
		DmRequest(self.gatewayobject).DMchannel(channel_id)

	def call(self, channelID, guildID=None, mute=False, deaf=False, video=False):
		MediaRequest(self.gatewayobject).call(channelID, guildID, mute, deaf, video)

	def endCall(self):
		MediaRequest(self.gatewayobject).endCall()


import json
import inspect

class LogLevel:
	INFO = '\033[94m'
	OK = '\033[92m'
	WARNING = '\033[93m'
	DEFAULT = '\033[m'

#The wrap that all http request api functions in discum use
class Wrapper:
	@staticmethod
	def sendRequest(reqsession, method, url, body=None, files=None, log=True):
		if hasattr(reqsession, method):
			stack = inspect.stack()
			function_name = "({}->{})".format(str(stack[1][0].f_locals['self']).split(' ')[0], stack[1][3])
			if log: #(sent) log msg
				sentMsg = '{} -> {}'.format(method.title(), url)
				print('{} [+] {} {}'.format(LogLevel.INFO, function_name, sentMsg))
				print(LogLevel.DEFAULT)
			data = {}
			if body != None:
				if isinstance(body, dict):
					data = {'data': json.dumps(body)}
				else:
					data = {'data': body}
				if log:
					bodyMsg = str(body)
					print('{} [+] {} {}'.format(LogLevel.INFO, function_name, bodyMsg))
					print(LogLevel.DEFAULT)
			if files != None:
				data['files'] = files
				if log:
					print('{} [+] {} {}'.format(LogLevel.INFO, function_name, "<file data>"))
					print(LogLevel.DEFAULT)
			response = getattr(reqsession, method)(url=url, **data)
			if log: #(received) log message
				receivedMsg = 'Response <- {}'.format(response.text)
				print('{} [+] {} {}'.format(LogLevel.OK, function_name, receivedMsg))
				print(LogLevel.DEFAULT)
			return response
		else:
			print('Invalid request method.')



#points to commands that help request info/actions using the gateway
#note, no need for importing GuildParse because resp is a Resp object (resp.parsed... does the trick)
#also, no need for importing GuildRequest because gatewayobj has that (self.gatewayobj.request... does the trick)

import time

class GuildCombo(object):
	def __init__(self, gatewayobj):
		self.gatewayobj = gatewayobj

	#fetchMembers helper function
	def reformat_member(self, memberdata, keep=[]): #memberdata comes in as a dict and leaves as a tuple (userID, memberdatareformatted). This is done to easily prevent duplicates in the member list when fetching.
		allProperties = ['pending', 'deaf', 'hoisted_role', 'presence', 'joined_at', 'public_flags', 'username', 'avatar', 'discriminator', 'premium_since', 'roles', 'is_pending', 'mute', 'nick', 'bot']
		if keep == None:
			remove = allProperties
		elif keep == "all":
			remove = []
		elif isinstance(keep, list):
			remove = list(set(allProperties) - set(keep))
		memberproperties = memberdata['member'] if 'member' in memberdata else memberdata
		userdata = memberproperties.pop('user', {})
		userID = userdata.pop('id', {})
		memberproperties.update(userdata)
		#filtering/removing
		if 'pending' in remove and 'pending' in memberproperties:
			del memberproperties['pending']
		if 'deaf' in remove and 'deaf' in memberproperties:
			del memberproperties['deaf']
		if 'hoisted_role' in remove and 'hoisted_role' in memberproperties:
			del memberproperties['hoisted_role']
		if 'presence' in remove and 'presence' in memberproperties:
			del memberproperties['presence']
		if 'joined_at' in remove and 'joined_at' in memberproperties:
			del memberproperties['joined_at']
		if 'public_flags' in remove and 'public_flags' in memberproperties:
			del memberproperties['public_flags']
		if 'username' in remove and 'username' in memberproperties:
			del memberproperties['username']
		if 'avatar' in remove and 'avatar' in memberproperties:
			del memberproperties['avatar']
		if 'discriminator' in remove and 'discriminator' in memberproperties:
			del memberproperties['discriminator']
		if 'premium_since' in remove and 'premium_since' in memberproperties:
			del memberproperties['premium_since']
		if 'roles' in remove and 'roles' in memberproperties:
			del memberproperties['roles']
		if 'is_pending' in remove and 'is_pending' in memberproperties:
			del memberproperties['is_pending']
		if 'mute' in remove and 'mute' in memberproperties:
			del memberproperties['mute']
		if 'nick' in remove and 'nick' in memberproperties:
			del memberproperties['nick']
		if 'bot' in remove and 'bot' in memberproperties:
			del memberproperties['bot']
		return userID, memberproperties

	#fetchMembers helper function
	def rangeCorrector(self, ranges): #just adds [0,99] at the beginning
		if [0,99] not in ranges:
			ranges.insert(0, [0,99])
		return ranges

	#fetchMembers helper function
	def getIndex(self, guild_id):
		return self.gatewayobj.memberFetchingStatus[guild_id][1]

	#fetchMembers helper function
	def getRanges(self, index, multiplier):
		initialNum = int(index*multiplier)
		return self.rangeCorrector([[initialNum, initialNum+99], [initialNum+100, initialNum+199]])

	#fetchMembers helper function
	def updateCurrent(self, guild_id):
		if not self.gatewayobj.finishedMemberFetching(guild_id): #yep still gotta check for this
			self.gatewayobj.memberFetchingStatus[guild_id][1] = self.gatewayobj.memberFetchingStatus[guild_id][0]+1

	#fetchMembers helper function
	def updatePrevious(self, guild_id):
		if not self.gatewayobj.finishedMemberFetching(guild_id):
			self.gatewayobj.memberFetchingStatus[guild_id][0] = self.gatewayobj.memberFetchingStatus[guild_id][1]

	def fetchMembers(self, resp, guild_id, channel_id, method, keep, considerUpdates, startIndex, stopIndex, reset, wait): #process is a little simpler than it looks. Keep in mind that there's no actual api endpoint to get members so this is a bit hacky. However, the method used below mimics how the official client loads the member list.
		if self.gatewayobj.READY:
			if self.gatewayobj.memberFetchingStatus.get(guild_id) == None: #request for lazy request
				self.gatewayobj.memberFetchingStatus[guild_id] = [startIndex, startIndex] #format is [previous index, current index]. This format is useful for the wait parameter.
				if not self.gatewayobj.session.guild(guild_id).hasMembers or reset:
					self.gatewayobj.session.guild(guild_id).resetMembers() #reset
				if len(self.gatewayobj.memberFetchingStatus["first"]) == 0:
					self.gatewayobj.memberFetchingStatus["first"] = [guild_id]
					self.gatewayobj.request.lazyGuild(guild_id, {channel_id: [[0,99]]}, typing=True, threads=False, activities=True, members=[])
				else:
					self.gatewayobj.request.lazyGuild(guild_id, {channel_id: [[0,99]]}, typing=True, activities=True)
			if self.gatewayobj.memberFetchingStatus.get(guild_id) != None and not self.gatewayobj.finishedMemberFetching(guild_id): #proceed with lazy requests
				index = self.getIndex(guild_id) #index always has the current value
				endFetching = False
				#find multiplier (this dictates the way the member list requested for)
				if method == "overlap": multiplier = 100
				elif method == "no overlap": multiplier = 200
				elif isinstance(method, int): multiplier = method
				elif isinstance(method, list) or isinstance(method, tuple): 
					if index<len(method):
						multiplier = method[index]
					else:
						endFetching = True #ends fetching right after resp parsed
				ranges = self.getRanges(index, multiplier) if not endFetching else [[0],[0]]
				#0th lazy request (separated from the rest because this happens "first")
				if index == startIndex and not self.gatewayobj.session.guild(guild_id).unavailable:
					self.updateCurrent(guild_id) #current = previous+1
					if wait!=None: time.sleep(wait)
					self.gatewayobj.request.lazyGuild(guild_id, {channel_id: ranges})
				elif resp.event.guild_member_list:
					parsed = resp.parsed.guild_member_list_update()
					if parsed['guild_id'] == guild_id and ('SYNC' in parsed['types'] or 'UPDATE' in parsed['types']):
						endFetching = False
						for ind,i in enumerate(parsed['types']):
							if i == 'SYNC':
								if len(parsed['updates'][ind]) == 0 and parsed['locations'][ind] in ranges[1:]: #checks if theres nothing in the SYNC data
									endFetching = True
								for item in parsed['updates'][ind]:
									if 'member' in item:
										member_id, member_properties = self.reformat_member(item, keep=keep)
										self.gatewayobj.session.guild(guild_id).updateOneMember(member_id, member_properties)
										if self.gatewayobj.log: print('<SYNC> updated member '+member_id)
								if not self.gatewayobj.finishedMemberFetching(guild_id) and (index-self.gatewayobj.memberFetchingStatus[guild_id][0])==1:
									if wait!=None: time.sleep(wait)
									self.updatePrevious(guild_id) #previous = current
							elif i == 'UPDATE' and considerUpdates: #this really only becomes useful for large guilds (because fetching members can take a quite some time for those guilds)
								for key in parsed['updates'][ind]:
									if key == 'member':
										member_id, member_properties = self.reformat_member(parsed['updates'][ind][key], keep=keep)
										self.gatewayobj.session.guild(guild_id).updateOneMember(member_id, member_properties)
										if self.gatewayobj.log: print('<UPDATE> updated member '+member_id)
							elif i == 'INVALIDATE':
								if parsed['locations'][ind] in ranges or parsed['member_count'] == 0:
									endFetching = True
						if ranges==[[0],[0]] or index>=stopIndex or ranges[-2][-1]>self.gatewayobj.session.guild(guild_id).memberCount or ranges[-1][-1]>self.gatewayobj.session.guild(guild_id).memberCount or endFetching: #putting whats most likely to happen first
							self.gatewayobj.memberFetchingStatus[guild_id] = "done"
							self.gatewayobj.removeCommand(
							    {
							        "function": self.fetchMembers,
							        "params": {
							            "guild_id": guild_id,
							            "channel_id": channel_id,
							            "method": method,
							            "keep": keep,
							            "considerUpdates": considerUpdates,
							            "startIndex": startIndex,
							            "stopIndex": stopIndex,
							            "reset": reset,
							            "wait": wait
							        },
							    }
							) #it's alright if you get a "not found in _after_message_hooks" error log. That's not an error for this situation.
						elif not self.gatewayobj.finishedMemberFetching(guild_id) and index==self.gatewayobj.memberFetchingStatus[guild_id][0]:
							self.updateCurrent(guild_id) #current = previous + 1
							self.gatewayobj.request.lazyGuild(guild_id, {channel_id: ranges})

	#the following 2 are test functions to show how the fetchMembers combo function works, might be removed in a later update
	def testfunc(self, resp):
		self.gatewayobj.removeCommand(self.testfunc)
		print('testfunc')
		pass

	def testfuncPOG(self, resp, pog):
		print('testfuncPOG')
		if pog:
			self.gatewayobj.removeCommand(self.testfuncPOG)
		pass



import filetype
import requests
import os

if __import__('sys').version.split(' ')[0] < '3.0.0':
    from urlparse import urlparse
else:
    from urllib.parse import urlparse

class Fileparse(object):
	def __init__(self, s, log): #s is the requests session object
		self.s = s
		self.log = log
		
	def parse(self, filelocation, isurl): #returns mimetype and extension if detected
		fd = b""
		if isurl:
			result = urlparse(filelocation)
			if all([result.scheme, result.netloc, result.path]): #if a link...
				fd = requests.get(filelocation,headers={"User-Agent":self.s.headers['User-Agent']},proxies=self.s.proxies).content
				kind = filetype.guess(fd)
				if kind is None:
					if self.log: print('Unsupported file type. Will attempt to send anyways.')
					return 'unsupported', 'unsupported', fd
				return kind.mime, kind.extension, fd
			else:
				if self.log: print('Invalid link.')
				return 'invalid', 'invalid', fd
		else:
			if os.path.isfile(filelocation):
				kind = filetype.guess(filelocation)
				if kind is None:
					if self.log: print('Unsupported file type. Will attempt to send anyways.')
					return 'unsupported', 'unsupported', fd
				return kind.mime, kind.extension, fd
			else:
				if self.log: print('Either not a file or file does not exist.')
				return 'invalid', 'invalid', fd



#parses response from gateway

#function names are just lowercase types, so for type GUILD_MEMBER_LIST_UPDATE, the function is guild_member_list_update
class GuildParse(object):
	@staticmethod
	def guild_member_list_update(response):
		memberdata = {
		    "online_count": response["d"]["online_count"],
		    "member_count": response["d"]["member_count"],
		    "id": response["d"]["id"],
		    "guild_id": response["d"]["guild_id"],
		    "hoisted_roles": response["d"]["groups"],
		    "types": [],
		    "locations": [],
		    "updates": []
		}
		#now time to look over ops
		for chunk in response['d']['ops']:
			memberdata['types'].append(chunk['op'])
			if chunk['op'] in ('SYNC', 'INVALIDATE'):
				memberdata['locations'].append(chunk['range'])
				if chunk['op'] == 'SYNC':
					memberdata['updates'].append(chunk['items'])
				else: #invalidate
					memberdata['updates'].append([])
			elif chunk['op'] in ('INSERT', 'UPDATE', 'DELETE'): #only update the 0,99 range btw
				memberdata['locations'].append(chunk['index'])
				if chunk['op'] == 'DELETE':
					memberdata['updates'].append([])
				else:
					memberdata['updates'].append(chunk['item'])
		#and that's it. mostly some renaming and shuffling around to help with usage and reading :)
		return memberdata
				




#handles building the superproperties dictionary. also includes getting the build number (since this is part of the super properties)

import ua_parser.user_agent_parser
import re
import time

class SuperProperties:
    '''
    https://luna.gitlab.io/discord-unofficial-docs/science.html#super-properties-object
    '''
    def __init__(self, s, buildnum="request", log=True):
        self.s = s
        self.buildnum = buildnum
        self.log = log

    def RequestBuildNumber(self):
        if self.log: print("Retrieving Discord's build number...")
        discord_login_page_exploration = self.s.get('https://discord.com/login').text
        time.sleep(1)
        try: #getting the build num is kinda experimental since who knows if discord will change where the build number is located...
            file_with_build_num = 'https://discord.com/assets/'+re.compile(r'assets/+([a-z0-9]+)\.js').findall(discord_login_page_exploration)[-2]+'.js' #fastest solution I could find since the last js file is huge in comparison to 2nd from last
            req_file_build = self.s.get(file_with_build_num).text
            index_of_build_num = req_file_build.find('buildNumber')+14
            discord_build_num = int(req_file_build[index_of_build_num:index_of_build_num+5])
            if self.log: print('Discord is currently on build number '+str(discord_build_num))
            return discord_build_num
        except:
            if self.log: print('Could not retrieve discord build number.')
            return None

    def GetSuperProperties(self, user_agent):
        parseduseragent = ua_parser.user_agent_parser.Parse(user_agent)
        browser_ver_list = [parseduseragent["user_agent"]["major"], parseduseragent["user_agent"]["minor"], parseduseragent["user_agent"]["patch"]]
        os_ver_list = [parseduseragent["os"]["major"], parseduseragent["os"]["minor"], parseduseragent["os"]["patch"]]
        sp = {
            "os": parseduseragent["os"]["family"],
            "browser": parseduseragent["user_agent"]["family"],
            "device": "",
            "browser_user_agent": parseduseragent["string"],
            "browser_version": ".".join(filter(None, browser_ver_list)),
            "os_version": ".".join(filter(None, os_ver_list)),
            "referrer": "",
            "referring_domain": "",
            "referrer_current": "",
            "referring_domain_current": "",
            "release_channel": "stable",
            "client_build_number": 73777,
            "client_event_source": None
        }
        if self.buildnum == "request":
            reqbuildnum = self.RequestBuildNumber()
            if reqbuildnum != None:
                sp["client_build_number"] = reqbuildnum
        elif isinstance(self.buildnum, int):
            sp["client_build_number"] = self.buildnum
        return sp



from ..RESTapiwrap import *

if __import__('sys').version.split(' ')[0] < '3.0.0':
    from urllib import quote
else:
    from urllib.parse import quote

class Guild(object):
	def __init__(self, discord, s, log): #s is the requests session object
		self.discord = discord
		self.s = s
		self.log = log

	'''
	invite codes / server info
	'''
	#get guild info from invite code
	def getInfoFromInviteCode(self,inviteCode):
		url = self.discord+"invites/"+inviteCode+"?with_counts=true"
		return Wrapper.sendRequest(self.s, 'get', url, log=self.log)

	#join guild with invite code
	def joinGuild(self,inviteCode):
		url = self.discord+"invites/"+inviteCode
		return Wrapper.sendRequest(self.s, 'post', url, log=self.log)

	def leaveGuild(self, guildID):
		url = self.discord+"users/@me/guilds/"+guildID
		return Wrapper.sendRequest(self.s, 'delete', url, log=self.log)

	def createInvite(self, channelID, max_age_seconds, max_uses, grantTempMembership, checkInvite, targetType): #has to be a channel thats in a guild. also checkInvite and targetType are basically useless.
		url = self.discord+"channels/"+channelID+"/invites"
		if max_age_seconds == False:
			max_age_seconds = 0
		if max_uses == False:
			max_uses = 0
		body = {"max_age": max_age_seconds, "max_uses": max_uses, "temporary": grantTempMembership}
		if checkInvite != "":
			body["validate"] = checkInvite
		if targetType != "":
			body["target_type"] = targetType
		return Wrapper.sendRequest(self.s, 'post', url, body, log=self.log)

	'''
	server moderation
	'''
	#kick a user
	def kick(self,guildID,userID,reason):
		url = self.discord+"guilds/%s/members/%s?reason=%s" % (guildID, userID, quote(reason))
		return Wrapper.sendRequest(self.s, 'delete', url, log=self.log)

	#ban a user
	def ban(self,guildID,userID,deleteMessagesDays,reason):
		url = self.discord+"guilds/%s/bans/%s" % (guildID, userID)
		body = {"delete_message_days": str(deleteMessagesDays), "reason": reason}
		return Wrapper.sendRequest(self.s, 'put', url, body, log=self.log)

	def revokeBan(self, guildID, userID):
		url = self.discord+"guilds/"+guildID+"/bans/"+userID
		return Wrapper.sendRequest(self.s, 'delete', url, log=self.log)



	#lookup a user in a guild. thx Echocage for finding this api endpoint
	'''Note, user clients do not run this api request, however it currently works without a problem. 
	Once the equivalent gateway command/request is added, the getGuildMember function will be removed from here.
	'''
	def getGuildMember(self, guildID, userID):
		url = self.discord+"/guilds/%s/members/%s" % (guildID, userID)
		return Wrapper.sendRequest(self.s, 'get', url, log=self.log)

	#get member verification data
	def getMemberVerificationData(self, guildID, with_guild=False, invite_code=None):
		url = "https://discord.com/api/v8/guilds/"+guildID+"/member-verification?with_guild="+str(with_guild).lower()
		if invite_code != None:
			url += "&invite_code="+invite_code
		return Wrapper.sendRequest(self.s, 'get', url, log=self.log)

	def agreeGuildRules(self, guildID, form_fields, version):
		url = "https://discord.com/api/v8/guilds/"+guildID+"/requests/@me"
		form_fields[0]['response'] = True
		body = {"version":version, "form_fields":json.dumps(form_fields)}
		return Wrapper.sendRequest(self.s, 'put', url, body, log=self.log)



#parses response from gateway
from .guild.parse import GuildParse
from .messages.parse import MessageParse

#function names are just lowercase types, so for type GUILD_MEMBER_LIST_UPDATE, the function is guild_member_list_update
class Parse(object):
	def __init__(self, response):
		self.response = response

	def auto(self): #auto parse, does not allow for custom inputs
		if hasattr(self, self.response['t'].lower()):
			return getattr(self, self.response['t'].lower())()
		return self.response['d'] #just return the value of d if there's no parse function for it yet

	def guild_member_list_update(self):
		return GuildParse.guild_member_list_update(self.response)

	def message_create(self):
		return MessageParse.message_create(self.response)



#request

class DmRequest(object):
	def __init__(self, gatewayobject):
		self.gatewayobject = gatewayobject

	def DMchannel(self, channel_id):
		self.gatewayobject.send({"op":self.gatewayobject.OPCODE.DM_UPDATE,"d":{"channel_id":channel_id}})


from ..RESTapiwrap import *
import time, datetime
import random

from .client_uuid import Client_UUID

class Science(object):
    def __init__(self, discord, s, log, analytics_token, userID, xfingerprint): #s is the requests session object
        self.discord = discord
        self.s = s
        self.log = log
        self.analytics_token = analytics_token
        self.xfingerprint = xfingerprint
        if userID == "0":
            date = datetime.datetime.now()
            unixts = time.mktime(date.timetuple())
            userID = str((int(unixts)*1000-1420070400000)*4194304)
        self.UUIDobj = Client_UUID(userID)

    def getCurrentUnixMS(self): #returns unix ts in milliseconds
        return int(time.mktime(datetime.datetime.now().timetuple()) * 1000)

    def getTrackingProperties(self, duration="random"):
        now = self.getCurrentUnixMS()
        trackingProperties = {"client_track_timestamp": now}
        if duration == "random":
            trackingProperties["client_send_timestamp"] = now+random.randint(40, 1000)
        else:
            trackingProperties["client_send_timestamp"] = now+duration
        trackingProperties["client_uuid"] = self.UUIDobj.calculate()
        return trackingProperties

    def science(self, events): #https://luna.gitlab.io/discord-unofficial-docs/science_events.html
        url = self.discord +"science"
        for event in events:
            if "type" not in event:
                event["type"] = "keyboard_shortcut_used" #random default type, will make an event helper later
            if "properties" not in event or "client_send_timestamp" not in event["properties"] or "client_track_timestamp" not in event["properties"] or "client_uuid" not in event["properties"]:
                event["properties"] = self.getTrackingProperties()
            else:
                self.UUIDobj.eventNum += 1
        body = {'token': self.analytics_token, 'events':events}
        if self.analytics_token == None: #if not logged in. ex: bot=discum.Client(token='poop')
            body.pop('token')
            import requests
            s = requests.Session()
            s.proxies.update(self.s.proxies)
            for header in self.s.headers:
                if header != "Authorization":
                    s.headers[header] = self.s.headers[header]
                else:
                    s.headers["Authorization"] = None
                    s.headers["X-fingerprint"] = self.xfingerprint
        else:
            s = self.s
        return Wrapper.sendRequest(s, 'post', url, body, log=self.log)


import websocket
import json
import time
import random
import base64
import zlib

if __import__('sys').version.split(' ')[0] < '3.0.0':
    import thread
else:
    import _thread as thread

from .session import session
from .response import Resp
from .request import Request

#other functions
from .guild.combo import GuildCombo

#gateway class
class GatewayServer:

    class LogLevel:
        SEND = '\033[94m'
        RECEIVE = '\033[92m'
        WARNING = '\033[93m'
        DEFAULT = '\033[m'

    class OPCODE:
        # Name                           Code    Client Action   Description
        DISPATCH =                       0  #    Receive         dispatches an event
        HEARTBEAT =                      1  #    Send/Receive    used for ping checking
        IDENTIFY =                       2  #    Send            used for client handshake
        PRESENCE_UPDATE =                3  #    Send            used to update the client status
        VOICE_STATE_UPDATE =             4  #    Send            used to join/move/leave voice channels
        VOICE_SERVER_PING =              5  #    Send            used for voice ping checking
        RESUME =                         6  #    Send            used to resume a closed connection
        RECONNECT =                      7  #    Receive         used to tell bots to reconnect...so..useless for discum
        REQUEST_GUILD_MEMBERS =          8  #    Send            used to request guild members (when searching for members in the search bar of a guild)
        INVALID_SESSION =                9  #    Receive         used to notify client they have an invalid session id
        HELLO =                          10 #    Receive         sent immediately after connecting, contains heartbeat and server debug information
        HEARTBEAT_ACK =                  11 #    Sent immediately following a client heartbeat that was received
        #GUILD_SYNC =                    12 #    Receive         supposedly guild_sync but not used...idk
        DM_UPDATE =                      13 #    Send            used to get dm features
        LAZY_REQUEST =                   14 #    Send            discord responds back with GUILD_MEMBER_LIST_UPDATE type SYNC...
        LOBBY_CONNECT =                  15 #    ??
        LOBBY_DISCONNECT =               16 #    ??
        LOBBY_VOICE_STATES_UPDATE =      17 #    Receive
        STREAM_CREATE =                  18 #    ??
        STREAM_DELETE =                  19 #    ??
        STREAM_WATCH =                   20 #    ??
        STREAM_PING =                    21 #    Send
        STREAM_SET_PAUSED =              22 #    ??
        REQUEST_APPLICATION_COMMANDS =   24 #    ??

    def __init__(self, websocketurl, token, super_properties, proxy_host=None, proxy_port=None, log=True):
        self.token = token
        self.super_properties = super_properties
        self.auth = {
                "token": self.token,
                "capabilities": 61,
                "properties": self.super_properties,
                "presence": {
                    "status": "online",
                    "since": 0,
                    "activities": [],
                    "afk": False
                },
                "compress": False,
                "client_state": {
                    "guild_hashes": {},
                    "highest_last_message_id": "0",
                    "read_state_version": 0,
                    "user_guild_settings_version": -1
                }
            }

        self.proxy_host = None if proxy_host in (None,False) else proxy_host
        self.proxy_port = None if proxy_port in (None,False) else proxy_port

        self.log = log

        self.interval = None
        self.session_id = None
        self.sequence = 0
        self.READY = False #becomes True once READY_SUPPLEMENTAL is received
        self.settings_ready = {}
        self.settings_ready_supp = {}

        #websocket.enableTrace(True) #for debugging
        self.ws = self._get_ws_app(websocketurl)

        self._after_message_hooks = []
        self._last_err = None

        self.connected = False
        self.resumable = False

        self.voice_data = {} #voice connections dependent on current (connected) session

        self.memberFetchingStatus = {"first": []}

        self.request = Request(self)

    #WebSocketApp, more info here: https://github.com/websocket-client/websocket-client/blob/master/websocket/_app.py#L79
    def _get_ws_app(self, websocketurl):
        sec_websocket_key = base64.b64encode(bytes(random.getrandbits(8) for _ in range(16))).decode() #https://websockets.readthedocs.io/en/stable/_modules/websockets/handshake.html
        headers = {
            "Accept-Encoding": "gzip, deflate, br",
            "Accept-Language": "en-US,en;q=0.9",
            "Cache-Control": "no-cache",
            "Connection": "Upgrade",
            "Host": "gateway.discord.gg",
            "Origin": "https://discord.com",
            "Pragma": "no-cache",
            "Sec-WebSocket-Extensions": "permessage-deflate; client_max_window_bits",
            "Sec-WebSocket-Key": sec_websocket_key,
            "Sec-WebSocket-Version": "13",
            "Upgrade": "websocket",
            "User-Agent": self.super_properties["browser_user_agent"]
        } #more info: https://stackoverflow.com/a/40675547

        ws = websocket.WebSocketApp(websocketurl,
                                    header = headers,
                                    on_open=lambda ws: self.on_open(ws),
                                    on_message=lambda ws, msg: self.on_message(ws, msg),
                                    on_error=lambda ws, msg: self.on_error(ws, msg),
                                    on_close=lambda ws: self.on_close(ws)
                                    )
        return ws

    def decompress(self, bmessage): #input is byte message
        data = self._zlib.decompress(bmessage)
        jsonmessage = json.loads(data.decode("UTF8"))
        return jsonmessage

    def on_open(self, ws):
        self.connected = True
        self.memberFetchingStatus = {"first": []}
        if self.log: print("Connected to websocket.")
        if not self.resumable:
            self.send({"op": self.OPCODE.IDENTIFY, "d": self.auth})
        else:
            self.resumable = False
            self.send({"op": self.OPCODE.RESUME, "d": {"token": self.token, "session_id": self.session_id, "seq": self.sequence-1 if self.sequence>0 else self.sequence}})

    def on_message(self, ws, message):
        self.sequence += 1
        response = self.decompress(message)
        if self.log: print('%s< %s%s' % (self.LogLevel.RECEIVE, response, self.LogLevel.DEFAULT))
        if response['op'] == self.OPCODE.HELLO: #only happens once, first message sent to client
            self.interval = (response["d"]["heartbeat_interval"]-2000)/1000
            thread.start_new_thread(self._heartbeat, ())
        elif response['op'] == self.OPCODE.INVALID_SESSION:
            if self.log: print("Invalid session.")
            if self.resumable:
                self.resumable = False
                self.sequence = 0
                self.close()
            else:
                self.sequence = 0
                self.close()
        if self.interval == None:
            if self.log: print("Identify failed.")
            self.close()
        if response['t'] == "READY":
            self.session_id = response['d']['session_id']
            self.settings_ready = response['d']
            self.session = session(self.settings_ready, {}) #so that you can parse session settings right after ready (if you dont listen for ready supp data)
        elif response['t'] == "READY_SUPPLEMENTAL":
            self.resumable = True #completely successful identify
            self.settings_ready_supp = response['d']
            self.session.settings_ready_supp = self.settings_ready_supp
            self.READY = True
        elif response['t'] == "GUILD_CREATE":
            guild_data = response['d']
            guild_data['members'] = {} #the member data originally stored in here is far from complete so might as well throw it away tbh and prepare for member fetching
            if response['d']['id'] in self.session.guildIDs:
                self.session.guild(response['d']['id']).setData(guild_data)
            else:
                self.session.add(guild_data)
        elif response['t'] in ("VOICE_SERVER_UPDATE", "VOICE_STATE_UPDATE"):
            self.voice_data.update(response['d']) #called twice, resulting in a dictionary with 12 keys
        resp = Resp(response)
        thread.start_new_thread(self._response_loop, (resp,))

    def on_error(self, ws, error):
        if self.log: print('%s%s%s' % (self.LogLevel.WARNING, error, self.LogLevel.DEFAULT))
        self._last_err = error

    def on_close(self, ws):
        self.connected = False
        self.READY = False #reset self.READY
        if self.log: print('websocket closed')

    #Discord needs heartbeats, or else connection will sever
    def _heartbeat(self):
        if self.log: print("entering heartbeat")
        while self.connected:
            time.sleep(self.interval)
            if not self.connected:
                break
            self.send({"op": self.OPCODE.HEARTBEAT,"d": self.sequence-1 if self.sequence>0 else self.sequence})

    #just a wrapper for ws.send
    def send(self, payload):
        if self.log: print('%s> %s%s' % (self.LogLevel.SEND, payload, self.LogLevel.DEFAULT))
        self.ws.send(json.dumps(payload))

    def close(self):
        self.connected = False
        self.READY = False #reset self.READY
        if self.log: print('websocket closed') #sometimes this message will print twice. Don't worry, that's not an error.
        self.ws.close()


    def command(self, func):
        if callable(func):
            self._after_message_hooks.append(func)
            return func
        elif isinstance(func, dict): #because I can't figure out out to neatly pass params to decorators :(. Normal behavior still works; use as usual.\
            priority = func.pop('priority', -1)
            self._after_message_hooks.insert(priority, func) #func here is a dict btw
            return func['function']

    #influenced by https://github.com/scrubjay55/Reddit_ChatBot_Python/blob/master/Reddit_ChatBot_Python/WebSockClient.py (Apache License 2.0)
    def _response_loop(self, resp):
        copy = self._after_message_hooks[:]
        for func in copy:
            if func in self._after_message_hooks:
                if callable(func): #just regular, input is a function
                    func(resp)
                elif isinstance(func, dict):
                    function = func['function']
                    params = func['params'] if 'params' in func else {}
                    function(resp, **params)

    def removeCommand(self, func):
        try:
            if callable(func):
                commandsCopy = [i['function'] if isinstance(i,dict) else i for i in self._after_message_hooks] #this contains a list of functions
                ind = commandsCopy.index(func)
                del self._after_message_hooks[ind]
            else: #assume type dict. value error if not found anyways
                self._after_message_hooks.remove(func)
        except ValueError:
            if self.log: print('%s not found in _after_message_hooks.' % func)
            pass

    def clearCommands(self):
        self._after_message_hooks = []

    def resetSession(self): #just resets some variables that in-turn, resets the session (client side). Do not run this while running run().
        self.interval = None
        self.session_id = None
        self.sequence = 0
        self.READY = False #becomes True once READY_SUPPLEMENTAL is received
        self.settings_ready = {}
        self.settings_ready_supp = {}
        self._last_err = None
        self.voice_data = {}
        self.resumable = False #you can't resume anyways without session_id and sequence

    #modified version of function run_4ever from https://github.com/scrubjay55/Reddit_ChatBot_Python/blob/master/Reddit_ChatBot_Python/Utils/WebSockClient.py (Apache License 2.0)
    def run(self, auto_reconnect=True):
        while auto_reconnect:
            self._zlib = zlib.decompressobj()
            self.ws.run_forever(ping_interval=10, ping_timeout=5, http_proxy_host=self.proxy_host, http_proxy_port=self.proxy_port)
            if isinstance(self._last_err, websocket._exceptions.WebSocketAddressException) or isinstance(self._last_err, websocket._exceptions.WebSocketTimeoutException):
                if self.resumable:
                    waitTime = random.randrange(1,6)
                    if self.log: print("Connection Dropped. Attempting to resume last valid session in %s seconds." % waitTime)
                    time.sleep(waitTime)
                else:
                    if self.log: print("Connection Dropped. Retrying in 10 seconds.")
                    time.sleep(10)
                continue
            elif not self.resumable: #this happens if you send an IDENTIFY but discord says INVALID_SESSION in response
                if self.log: print("Connection Dropped. Retrying in 10 seconds.")
                time.sleep(10)
                continue
            else:
                self.resumable = True
                return 0
        if not auto_reconnect:
            self._zlib = zlib.decompressobj()
            self.ws.run_forever(ping_interval=10, ping_timeout=5, http_proxy_host=self.proxy_host, http_proxy_port=self.proxy_port)

##########################################################

    '''
    Guild/Server stuff
    '''

    def getMemberFetchingParams(self, targetRangeStarts): #more for just proof of concept. targetRangeStarts must not contain duplicates and must be a list of integers
        targetRangeStarts = {i:1 for i in targetRangeStarts} #remove duplicates but preserve order
        if targetRangeStarts.get(0)!=None and targetRangeStarts.get(100)!=None:
            keys = list(targetRangeStarts)
            if keys.index(100)<keys.index(0):
                targetRangeStarts.pop(0) #needs to be removed or else fetchMembers will enter an infinite loop because of how discord responds to member list requests
        startIndex = 1 #can't start at 0 because can't divide by 0. No need to specify a stop index since fetchMembers continues until end of multipliers
        method = [0] #because startIndex is 1
        for index,i in enumerate(targetRangeStarts):
            method.append(i/(index+1))
        return startIndex, method #return startIndex and multipliers

    def fetchMembers(self, guild_id, channel_id, method="overlap", keep=[], considerUpdates=True, startIndex=0, stopIndex=1000000000, reset=True, wait=None, priority=0):
        if guild_id in self.memberFetchingStatus:
            del self.memberFetchingStatus[guild_id] #just resetting tracker on the specific guild_id
        self.command(
            {
                "function": GuildCombo(self).fetchMembers,
                "priority": priority,
                "params": {
                    "guild_id": guild_id,
                    "channel_id": channel_id,
                    "method": method,
                    "keep": keep,
                    "considerUpdates": considerUpdates,
                    "startIndex": startIndex,
                    "stopIndex": stopIndex,
                    "reset": reset,
                    "wait": wait
                },
            }
        )


    def finishedMemberFetching(self, guild_id):
        return self.memberFetchingStatus.get(guild_id) == "done"

    '''
    test stuff (these show how to add combo functions)
    '''
    def testfunc(self):
        self.command({'function': GuildCombo(self).testfunc, 'priority': 0})

    def testfuncPOG(self, pog):
        self.command({'function': GuildCombo(self).testfuncPOG, 'priority': 0, 'params': {'pog': pog}})



from ..RESTapiwrap import *
from .totp import TOTP
import requests
import time

class Login:
    '''
    Manages HTTP authentication
    '''
    def __init__(self, s, discordurl, log):
        self.s = s
        self.discord = discordurl
        self.log = log
        if 'Authorization' in s.headers or 'X-Fingerprint' in s.headers:
            self.editedS = requests.Session()
            self.editedS.headers.update(s.headers)
            if 'Authorization' in self.editedS.headers:
                del self.editedS.headers['Authorization']
            if 'X-Fingerprint' in self.editedS.headers:
                del self.editedS.headers['X-Fingerprint']
            self.editedS.proxies.update(s.proxies)
        else:
            self.editedS = s

    def GetXFingerprint(self):
        url = self.discord + "experiments"
        reqxfinger = Wrapper.sendRequest(self.editedS, 'get', url, log=self.log)
        xfingerprint = json.loads(reqxfinger.content).get('fingerprint')
        if self.log:
            if xfingerprint == None:
                print('xfingerprint could not be fetched.')
        return xfingerprint

    def GetToken(self, email, password, undelete=False, captcha=None, source=None, gift_code_sku_id=None, secret="", code=""):
        url = self.discord + "auth/login"
        self.xfingerprint = self.GetXFingerprint()
        self.s.headers.update({"X-Fingerprint": self.xfingerprint})
        self.editedS.headers.update({"X-Fingerprint": self.xfingerprint})
        body = {
            "email": email,
            "password": password,
            "undelete": undelete,
            "captcha_key": captcha,
            "login_source": source,
            "gift_code_sku_id": gift_code_sku_id
        }
        response = Wrapper.sendRequest(self.editedS, 'post', url, body, log=self.log)
        result = json.loads(response.content)
        if 'mfa' in result and 'sms' in result: #sms login not implemented yet
            if result['mfa'] == True and result['sms'] == False:
                time.sleep(2) #2 seconds is minimal, don't want to look too automated
                ticket = result['ticket']
                if secret != "":
                    code = TOTP(secret).generateTOTP()
                code = str(code) #just in case an int is inputted
                totpUrl = self.discord+"auth/mfa/totp"
                totpBody = {
                    "code": code,
                    "ticket": ticket,
                    "login_source": source,
                    "gift_code_sku_id": gift_code_sku_id
                }
                totpResponse = Wrapper.sendRequest(self.editedS, 'post', totpUrl, totpBody, log=self.log)
                result = json.loads(totpResponse.content)
        self.__token = result['token']
        return self.__token, self.xfingerprint



class Embedder(object):
    def __init__(self):
        self.jsonEmbed = {"fields": []} #Create a 'fields' key otherwise we can't add anything to it, plus it doesn't effect anything if nothing is added.

    def read(self):
        return self.jsonEmbed

    def title(self,title):
        self.jsonEmbed.update({'title': title}) 

    def description(self,description):
        self.jsonEmbed.update({'description': description}) 

    def url(self,url):
        self.jsonEmbed.update({'url': url}) 

    def color(self,color):
        self.jsonEmbed.update({'color': color}) 

    def footer(self,text,iconURL=""):
        self.jsonEmbed.update({'footer': {
            'icon_url': iconURL,
            'text': text
        }}) 

    def image(self,url):
        self.jsonEmbed.update({'image': {
            'url': url,
        }})

    def thumbnail(self,url):
        self.jsonEmbed.update({'thumbnail': {
            'url': url,
        }})

    def author(self,name,url="",icon_url=""):
        self.jsonEmbed.update({'author': {
            'name': name,
            'url': url,
            'icon_url':icon_url
        }})

    def fields(self,name,value,inline=False):
        self.jsonEmbed['fields'].append({
            'name': name,
            'value': value,
            'inline': inline
        })
    
    


class session:
    settings_ready = {}
    settings_ready_supp = {}
    def __init__(self, input_settings_ready, input_settings_ready_supp):
        session.settings_ready = input_settings_ready
        session.settings_ready_supp = input_settings_ready_supp
        self.guild = guild
        self.DM = DM
        self.relationship = relationship
        self.userGuildSetting = userGuildSetting

    def read(self): #returns all session settings
        return [self.settings_ready, self.settings_ready_supp]

    ###***USER***###
    @property
    def user(self):
        return self.settings_ready['user']

    ###***GUILDS***### (general)
    @property
    def guilds(self):
        return self.settings_ready['guilds']

    @property
    def guildIDs(self):
        return [self.settings_ready['guilds'][i]['id'] for i in range(len(self.settings_ready['guilds']))]

    @property
    def mergedPresences(self):
        return self.settings_ready_supp['merged_presences']['guilds']

    @property
    def positions(self): #outputs your roles, nick, joined at, etc for all guilds you're in
        return self.settings_ready['merged_members']

    def add(self, guilddata):
        self.settings_ready['guilds'].append(guilddata)

    ###***RELATIONSHIPS***### (general)
    @property
    def relationships(self):
        return self.settings_ready['relationships']
    
    @property
    def relationshipIDs(self):
        return [self.relationships[i]['id'] for i in range(len(self.relationships))]

    #friends
    @property
    def friends(self):
        f = []
        for i in range(len(self.relationships)):
            if self.relationships[i]['type'] == 1:
                f.append(self.relationships[i])
        return f
    
    @property
    def friendIDs(self):
        return [self.friends[i]['user_id'] for i in range(len(self.friends))]

    #blocked    
    @property
    def blocked(self):
        b = []
        for i in range(len(self.relationships)):
            if self.relationships[i]['type'] == 2:
                b.append(self.relationships[i])
        return b
    
    @property
    def blockedIDs(self):
        return [self.blocked[i]['user_id'] for i in range(len(self.blocked))]
    
    #incoming    
    @property
    def incomingFriendRequests(self):
        ifr = []
        for i in range(len(self.relationships)):
            if self.relationships[i]['type'] == 3:
                ifr.append(self.relationships[i])
        return ifr
    
    @property
    def incomingFriendRequestIDs(self):
        return [self.incomingFriendRequests[i]['user_id'] for i in range(len(self.incomingFriendRequests))]

    #outgoing
    @property
    def outgoingFriendRequests(self):
        ofr = []
        for i in range(len(self.relationships)):
            if self.relationships[i]['type'] == 4:
                ofr.append(self.relationships[i])
        return ofr
    
    @property
    def outgoingFriendRequestIDs(self):
        return [self.outgoingFriendRequests[i]['user_id'] for i in range(len(self.outgoingFriendRequests))]

    #friend merged presences    
    @property
    def allFriendMergedPresences(self):
        return self.settings_ready_supp['merged_presences']['friends']
    
    @property
    def allFriendMergedPresenceIDs(self):
        return [self.allFriendMergedPresences[i]['user_id'] for i in range(len(self.allFriendMergedPresences))]
        

    ###***DMs***### (general)
    @property
    def DMs(self):
        return self.settings_ready['private_channels']

    @property
    def DMIDs(self):
        return [self.DMs[i]['id'] for i in range(len(self.DMs))] #discord sometimes calls these channel IDs...
        

    ###***USER SETTINGS***### (general)
    @property
    def userGuildSettings(self):
        return self.settings_ready['user_guild_settings'] #personal settings for a server, like whether or not you want @everyone pings to show
    
    @property
    def userSettings(self):
        return self.settings_ready['user_settings']
    
    @property
    def optionsForUserSettings(self):
        return list(self.settings_ready['user_settings'].keys())
        

    ###other stuff
    @property
    def mergedPresences(self):
        return self.settings_ready_supp['merged_presences'] #includes both guild and friend presences, but only the users that are not offline

    @property
    def analyticsToken(self):
        return self.settings_ready['analytics_token']

    @property
    def connectedAccounts(self):
        return self.settings_ready['connected_accounts']

    @property
    def consents(self):
        return self.settings_ready['consents']

    @property
    def experiments(self):
        return self.settings_ready['experiments']

    @property
    def friendSuggestionCount(self):
        return self.settings_ready['friend_suggestion_count']

    @property
    def guildExperiments(self):
        return self.settings_ready['guild_experiments']

    @property
    def readStates(self):
        return self.settings_ready['read_state'] #another advantage of using websockets instead of requests (see https://github.com/discord/discord-api-docs/issues/13)

    @property
    def geoOrderedRtcRegions(self):
        return self.settings_ready['geo_ordered_rtc_regions']

    @property
    def cachedUsers(self):
        return self.settings_ready['users']

    @property
    def tutorial(self):
        return self.settings_ready['tutorial'] #tutorial on what? guess we'll never know...ask discord maybe?


###specific guild
class guild(session):
    def __init__(self, guildID):
        self.guildIndex = None
        for i in range(len(session.settings_ready['guilds'])):
            if session.settings_ready['guilds'][i]['id'] == guildID:
                self.guildIndex = i

    @property
    def data(self):
        return session.settings_ready['guilds'][self.guildIndex]

    def setData(self, newData):
        session.settings_ready['guilds'][self.guildIndex] = newData

    def modify(self, modifications):
        session.settings_ready['guilds'][self.guildIndex].update(modifications)

    @property
    def unavailable(self):
        return 'unavailable' in session.settings_ready['guilds'][self.guildIndex]

    @property
    def hasMembers(self):
        return 'members' in session.settings_ready['guilds'][self.guildIndex]

    @property
    def members(self):
        return session.settings_ready['guilds'][self.guildIndex]['members']

    def resetMembers(self):
        session.settings_ready['guilds'][self.guildIndex]['members'] = {}

    def updateOneMember(self, userID, userProperties):
        session.settings_ready['guilds'][self.guildIndex]['members'][userID] = userProperties

    def updateMembers(self, memberdata): #where member data is a dictionary --> {userId: {properties}}
        session.settings_ready['guilds'][self.guildIndex]['members'].update(memberdata)

    @property
    def owner(self):
        return session.settings_ready['guilds'][self.guildIndex]['owner_id'] #returns type int

    @property
    def boostLvl(self):
        return session.settings_ready['guilds'][self.guildIndex]['premium_tier']

    @property
    def emojis(self):
        return session.settings_ready['guilds'][self.guildIndex]['emojis']

    @property
    def banner(self):
        return session.settings_ready['guilds'][self.guildIndex]['banner'] #if returns 'nil' then there's no banner

    @property
    def discoverySplash(self): #not sure what this is about, something about server discoverability i guess (https://discord.com/developers/docs/resources/guild)
        return session.settings_ready['guilds'][self.guildIndex]['discovery_splash']

    @property
    def msgNotificationSettings(self): #returns an int, 0=all messages, 1=only mentions (https://discord.com/developers/docs/resources/guild#guild-object-default-message-notification-level)
        return session.settings_ready['guilds'][self.guildIndex]['default_message_notifications']

    @property
    def rulesChannelID(self): #nil if no rules channel id, idk if this always works so it might actually be more useful just to look for the word "rules" in channel names
        return session.settings_ready['guilds'][self.guildIndex]['rules_channel_id']

    @property
    def verificationLvl(self): #returns an int, 0-4 (https://discord.com/developers/docs/resources/guild#guild-object-verification-level)
        return session.settings_ready['guilds'][self.guildIndex]['verification_level']

    @property
    def features(self): #returns a list of strings (https://discord.com/developers/docs/resources/guild#guild-object-guild-features)
        return session.settings_ready['guilds'][self.guildIndex]['features']

    @property
    def joinTime(self): #returns when you joined the server, type string
        return session.settings_ready['guilds'][self.guildIndex]['joined_at']

    @property
    def region(self):
        return session.settings_ready['guilds'][self.guildIndex]['region']

    @property
    def applicationID(self): #returns application id of the guild creator if it is bot-created (https://discord.com/developers/docs/resources/guild#guild-object-guild-features)
        return session.settings_ready['guilds'][self.guildIndex]['application_id']

    @property
    def afkChannelID(self): #not sure what this is
        return session.settings_ready['guilds'][self.guildIndex]['afk_channel_id']

    @property
    def icon(self): #https://discord.com/developers/docs/reference#image-formatting
        return session.settings_ready['guilds'][self.guildIndex]['icon']

    @property
    def name(self):
        return session.settings_ready['guilds'][self.guildIndex]['name']

    @property
    def maxVideoChannelUsers(self):
        return session.settings_ready['guilds'][self.guildIndex]['max_video_channel_users']

    @property
    def roles(self): #https://discord.com/developers/docs/topics/permissions#role-object
        return session.settings_ready['guilds'][self.guildIndex]['roles']

    @property
    def publicUpdatesChannelID(self):
        return session.settings_ready['guilds'][self.guildIndex]['public_updates_channel_id']

    @property
    def systemChannelFlags(self): #https://discord.com/developers/docs/resources/guild#guild-object-system-channel-flags
        return session.settings_ready['guilds'][self.guildIndex]['system_channel_flags']

    @property
    def mfaLvl(self): #https://discord.com/developers/docs/resources/guild#guild-object-mfa-level
        return session.settings_ready['guilds'][self.guildIndex]['mfa_level']

    @property
    def afkTimeout(self): #returns type int, unit seconds, https://discord.com/developers/docs/resources/guild
        return session.settings_ready['guilds'][self.guildIndex]['afk_timeout']

    @property
    def hashes(self): #https://github.com/discord/discord-api-docs/issues/1642
        return session.settings_ready['guilds'][self.guildIndex]['guild_hashes']

    @property
    def systemChannelID(self): #returns an int, the id of the channel where guild notices such as welcome messages and boost events are posted
        return session.settings_ready['guilds'][self.guildIndex]['system_channel_id']

    @property
    def lazy(self): #slightly different naming format since it returns a boolean (https://luna.gitlab.io/discord-unofficial-docs/lazy_guilds.html)
        if session.settings_ready['guilds'][self.guildIndex]['lazy'] == 'true':
            return True
        else:
            return False

    @property
    def numBoosts(self): #get number of boosts the server has gotten
        return session.settings_ready['guilds'][self.guildIndex]['premium_subscription_count']

    @property
    def large(self): #slightly different naming format since it returns a boolean, large if more than 250 members
        if session.settings_ready['guilds'][self.guildIndex]['large'] == 'true':
            return True
        else:
            return False

    @property
    def explicitContentFilter(self): #https://discord.com/developers/docs/resources/guild#guild-object-explicit-content-filter-level
        return session.settings_ready['guilds'][self.guildIndex]['explicit_content_filter']

    @property
    def splashHash(self): #not sure what this is for
        return session.settings_ready['guilds'][self.guildIndex]['splash']

    @property
    def memberCount(self):
        return session.settings_ready['guilds'][self.guildIndex]['member_count']

    @property
    def description(self):
        return session.settings_ready['guilds'][self.guildIndex]['description']

    @property
    def vanityUrlCode(self):
        return session.settings_ready['guilds'][self.guildIndex]['vanity_url_code']

    @property
    def preferredLocale(self):
        return session.settings_ready['guilds'][self.guildIndex]['preferred_locale']

    @property
    def allChannels(self): #returns all categories and all channels, all the data about that, wall of data so it can be a bit overwhelming, useful if you want to check how many channels your server has since discord counts categories as channels
        return session.settings_ready['guilds'][self.guildIndex]['channels']

    @property
    def categories(self): #all data about all guild categories, can be overwhelming
        all_channels = session.settings_ready['guilds'][self.guildIndex]['channels']
        all_categories = []
        for channel in all_channels: #https://discord.com/developers/docs/resources/channel#channel-object-channel-types
            if channel['type'] == 4:
                all_categories.append(channel)
        return all_categories

    @property
    def categoryIDs(self):
        return [self.categories[i]['id'] for i in range(len(self.categories))]

    def categoryData(self,categoryID):
        for i in range(len(self.categories)):
            if self.categories[i]['id'] == categoryID:
                return self.categories[i]
        return None #category not found

    @property
    def channels(self): #all data about all guild channels, can be overwhelming
        all_channels = session.settings_ready['guilds'][self.guildIndex]['channels']
        all_non_categories = []
        for channel in all_channels: #https://discord.com/developers/docs/resources/channel#channel-object-channel-types
            if channel['type'] != 4:
                all_non_categories.append(channel)
        return all_non_categories

    @property
    def channelIDs(self):
        return [self.channels[i]['id'] for i in range(len(self.channels))]

    def channelData(self,channelID):
        for i in range(len(self.channels)):
            if self.channels[i]['id'] == channelID:
                return self.channels[i]
        return None #channel not found

    @property
    def voiceStates(self): #https://discord.com/developers/docs/resources/voice#voice-state-object
        return session.settings_ready_supp['guilds'][self.guildIndex]['voice_states']

    @property
    def notOfflineCachedMembers(self): #so, not offline, but not all the not-offline members so....not offline cached members. I know I know long function name, but it's descriptive.
        return session.settings_ready_supp['merged_members'][self.guildIndex]

    @property
    def notOfflineCachedMemberIDs(self):
        return [self.notOfflineCachedMembers[i]['user_id'] for i in range(len(self.notOfflineCachedMembers))]

    def notOfflineCachedMemberData(self,userID):
        for i in range(len(self.notOfflineCachedMembers)):
            if self.notOfflineCachedMembers[i]['user_id'] == userID:
                return self.notOfflineCachedMembers[i]
        return None

    @property
    def mergedPresences(self): #this is different from getGuildNotOfflineCachedMembers because you can see presence data. also, not sure why, but a diff number of members shows up here than getGuildNotOfflineCachedMembers
        return session.settings_ready_supp['merged_presences']['guilds'][self.guildIndex]

    @property
    def mergedPresenceIDs(self): #returns user ids in a list
        return [self.mergedPresences[i]['user_id'] for i in range(len(self.mergedPresences))]

    def mergedPresenceData(self,userID):
        for i in range(len(self.mergedPresences)):
            if self.mergedPresences[i]['user_id'] == userID:
                return self.mergedPresences[i]
        return None

    @property
    def position(self): #my roles in guild and other data
        return session.settings_ready['merged_members'][self.guildIndex]

###specific relationship
class relationship(session): #not the same organization as class guild
    def __init__(self, userID):
        self.userID = userID

    @property
    def data(self): #usernames and discriminators are no longer provided in this data
        for i in range(len(session.settings_ready['relationships'])):
            if session.settings_ready['relationships'][i]['user_id'] == self.userID:
                return session.settings_ready['relationships'][i]
        return None

    @property
    def friendMergedPresenceData(self):
        for i in range(len(session.settings_ready_supp['merged_presences']['friends'])):
            if session.settings_ready_supp['merged_presences']['friends'][i]['user_id'] == self.userID:
                return session.settings_ready_supp['merged_presences']['friends'][i]
        return None

###specific DM
class DM(session):
    def __init__(self, DMID):
        self.DMID = DMID

    @property
    def data(self):
        for i in range(len(session.settings_ready['private_channels'])):
            if session.settings_ready['private_channels'][i]['id'] == self.DMID:
                return session.settings_ready['private_channels'][i]
        return None

    @property
    def recipients(self): #returns everyone in that DM except you
        return self.data['recipient_ids']

###specific User Guild Settings
class userGuildSetting(session):
    def __init__(self, guildID):
        self.guildID = guildID

    @property
    def data(self):
        if len(session.settings_ready['user_guild_settings']['entries']) == 0:
            return None
        for i in range(len(session.settings_ready['user_guild_settings']['entries'])):
            if session.settings_ready['user_guild_settings']['entries'][i]['guild_id'] == self.guildID:
                return session.settings_ready['user_guild_settings']['entries'][i]
        return None



import base64
from ..RESTapiwrap import *

class User(object):
	def __init__(self, discord, s, log): #s is the requests session object
		self.discord = discord
		self.s = s
		self.log = log
		
	#def getDMs(self): #websockets does this now
	#	url = self.discord+"users/@me/channels"
	#	return self.s.get(url)

	#def getGuilds(self): #websockets does this now
	#	url = self.discord+"users/@me/guilds"
	#	return self.s.get(url)

	#def getRelationships(self): #websockets does this now
	#	url = self.discord+"users/@me/relationships"
	#	return self.s.get(url)

	def requestFriend(self,user):
		if "#" in user:
			url = self.discord+"users/@me/relationships"
			body = {"username": user.split("#")[0], "discriminator": int(user.split("#")[1])}
			return Wrapper.sendRequest(self.s, 'post', url, body, log=self.log)
		else:
			url = self.discord+"users/@me/relationships/"+user
			body = {}
			return Wrapper.sendRequest(self.s, 'put', url, body, log=self.log)

	def acceptFriend(self,userID):
		url = self.discord+"users/@me/relationships/"+userID
		body = {}
		return Wrapper.sendRequest(self.s, 'put', url, body, log=self.log)

	def removeRelationship(self,userID): #for removing friends, unblocking people
		url = self.discord+"users/@me/relationships/"+userID
		return Wrapper.sendRequest(self.s, 'delete', url, log=self.log)

	def blockUser(self,userID):
		url = self.discord+"users/@me/relationships/"+userID
		body = {"type": 2}
		return Wrapper.sendRequest(self.s, 'put', url, body, log=self.log)

	def getProfile(self,userID):
		url = self.discord+"users/"+userID+"/profile"
		return Wrapper.sendRequest(self.s, 'get', url, log=self.log)

	def info(self, with_analytics_token=None): #simple. bot.info() for own user data
		url = self.discord+"users/@me"
		if with_analytics_token!=None:
			with_analytics_token = str(with_analytics_token).lower()
			url += "?with_analytics_token="+with_analytics_token
		return Wrapper.sendRequest(self.s, 'get', url, log=self.log)

	def getUserAffinities(self):
		url = self.discord+"users/@me/affinities/users"
		return Wrapper.sendRequest(self.s, 'get', url, log=self.log)

	#guild affinities with respect to current user, decided to organize this wrap in here b/c that's how the api is organized
	def getGuildAffinities(self):
		url = self.discord+"users/@me/affinities/guilds"
		return Wrapper.sendRequest(self.s, 'get', url, log=self.log)

	def getMentions(self, limit=25, roleMentions=True, everyoneMentions=True):
		roleMentions = str(roleMentions).lower()
		everyoneMentions = str(everyoneMentions).lower()
		url = self.discord+"users/@me/mentions?limit="+str(limit)+"&roles="+roleMentions+"&everyone="+everyoneMentions
		return Wrapper.sendRequest(self.s, 'get', url, log=self.log)

	def removeMentionFromInbox(self, messageID):
		url = self.discord+"users/@me/mentions/"+messageID
		return Wrapper.sendRequest(self.s, 'delete', url, log=self.log)

	def getMyStickers(self):
		url = self.discord+"users/@me/sticker-packs"
		return Wrapper.sendRequest(self.s, 'get', url, log=self.log)

	'''
	Profile Edits
	'''	
	def setStatus(self,status): #status options are: online, idle, dnd, invisible
		url = self.discord+"users/@me/settings"
		if status in ("online", "idle", "dnd", "invisible"):
			body = {"status": status}
		elif status in ('', None):
			body = {"custom_status": None}
		else:
			body = {"custom_status":{"text": str(status)}}
		return Wrapper.sendRequest(self.s, 'patch', url, body, log=self.log)

	def setAvatar(self, imagePath): #local image
		url = self.discord+"users/@me"
		with open(imagePath, "rb") as image:
			encodedImage = base64.b64encode(image.read()).decode('utf-8')
		body = {"avatar":"data:image/png;base64,"+encodedImage}
		return Wrapper.sendRequest(self.s, 'patch', url, body, log=self.log)

	def setUsername(self, username, password):
		url = self.discord+"users/@me"
		body = {"username": username, "password": password}
		return Wrapper.sendRequest(self.s, 'patch', url, body, log=self.log)

	def setEmail(self, email, password):
		url = self.discord+"users/@me"
		body = {"email": email, "password": password}
		return Wrapper.sendRequest(self.s, 'patch', url, body, log=self.log)

	def setPassword(self, new_password, password):
		url = self.discord+"users/@me"
		body = {"password": password, "new_password": new_password}
		return Wrapper.sendRequest(self.s, 'patch', url, body, log=self.log)

	def setDiscriminator(self, discriminator, password):
		url = self.discord+"users/@me"
		body = {"password":password, "discriminator":discriminator}
		return Wrapper.sendRequest(self.s, 'patch', url, body, log=self.log)
	'''
	More settings stuff
	'''
	def setHypesquad(self, house):
		url = self.discord+"hypesquad/online"
		if house.lower() == "bravery":
			body = {"house_id": 1}
		elif house.lower() == "brilliance":
			body = {"house_id": 2}
		elif house.lower() == "balance":
			body = {"house_id": 3}
		return Wrapper.sendRequest(self.s, 'post', url, body, log=self.log)

	def leaveHypesquad(self):
		url = self.discord+"hypesquad/online"
		return Wrapper.sendRequest(self.s, 'delete', url, log=self.log)

	def setLocale(self, locale):
		url = self.discord+"users/@me/settings"
		body = {"locale": locale}
		return Wrapper.sendRequest(self.s, 'patch', url, body, log=self.log)

	def enable2FA(self, code, secret, password): #returns new token plus backup codes
		url = self.discord+"users/@me/mfa/totp/enable"
		body = {"code": code, "secret": secret, "password": password}
		return Wrapper.sendRequest(self.s, 'post', url, body, log=self.log)

	def disable2FA(self, code):
		url = self.discord+"users/@me/mfa/totp/disable"
		body = {"code": code}
		return Wrapper.sendRequest(self.s, 'post', url, body, log=self.log)

	def getRTCregions(self):
		url = "https://latency.discord.media/rtc"
		return Wrapper.sendRequest(self.s, 'get', url, log=self.log)

	def setAFKtimeout(self, timeout_seconds):
		url = self.discord+"users/@me/settings"
		body = {"afk_timeout": timeout_seconds}
		return Wrapper.sendRequest(self.s, 'patch', url, body, log=self.log)

	def setTheme(self, theme):
		url = self.discord+"users/@me/settings"
		body = {"theme": theme.lower()}
		return Wrapper.sendRequest(self.s, 'patch', url, body, log=self.log)

	def setMessageDisplay(self, CozyOrCompact):
		url = self.discord+"users/@me/settings"
		if CozyOrCompact.lower() == "compact":
			body = {"message_display_compact": True}
		else:
			body = {"message_display_compact": False}
		return Wrapper.sendRequest(self.s, 'patch', url, body, log=self.log)

	def enableDevMode(self, enable):
		url = self.discord+"users/@me/settings"
		body = {"developer_mode": enable}
		return Wrapper.sendRequest(self.s, 'patch', url, body, log=self.log)

	def activateApplicationTestMode(self, applicationID):
		url = self.discord+"applications/"+applicationID+"/skus"
		return Wrapper.sendRequest(self.s, 'get', url, log=self.log)

	def getApplicationData(self, applicationID, with_guild=False):
		url = self.discord+"applications/"+applicationID+"/public?with_guild="+str(with_guild).lower()
		return Wrapper.sendRequest(self.s, 'get', url, log=self.log)

	def getBackupCodes(self, password, regenerate=False):
		url = self.discord+"users/@me/mfa/codes"
		body = {"password": password, "regenerate": regenerate}
		return Wrapper.sendRequest(self.s, 'post', url, body, log=self.log)

	'''
	App Settings - Text&Images
	'''
	def enableInlineMedia(self, enable):
		url = self.discord+"users/@me/settings"
		body = {"inline_embed_media": enable}
		return Wrapper.sendRequest(self.s, 'patch', url, body, log=self.log)

	def enableLargeImagePreview(self, enable):
		url = self.discord+"users/@me/settings"
		body = {"inline_attachment_media": enable}
		return Wrapper.sendRequest(self.s, 'patch', url, body, log=self.log)

	def enableGifAutoPlay(self, enable):
		url = self.discord+"users/@me/settings"
		body = {"gif_auto_play": enable}
		return Wrapper.sendRequest(self.s, 'patch', url, body, log=self.log)

	def enableLinkPreview(self, enable):
		url = self.discord+"users/@me/settings"
		body = {"render_embeds": enable}
		return Wrapper.sendRequest(self.s, 'patch', url, body, log=self.log)

	def enableReactionRendering(self, enable):
		url = self.discord+"users/@me/settings"
		body = {"render_reactions": enable}
		return Wrapper.sendRequest(self.s, 'patch', url, body, log=self.log)

	def enableAnimatedEmoji(self, enable):
		url = self.discord+"users/@me/settings"
		body = {"animate_emoji": enable}
		return Wrapper.sendRequest(self.s, 'patch', url, body, log=self.log)

	def enableEmoticonConversion(self, enable):
		url = self.discord+"users/@me/settings"
		body = {"convert_emoticons": enable}
		return Wrapper.sendRequest(self.s, 'patch', url, body, log=self.log)

	def setStickerAnimation(self, setting):
		url = self.discord+"users/@me/settings"
		if setting.lower() == "always":
			body = {"animate_stickers": 0}
		elif setting.lower() == "interaction":
			body = {"animate_stickers": 1}
		elif setting.lower() == "never":
			body = {"animate_stickers": 2}
		return Wrapper.sendRequest(self.s, 'patch', url, body, log=self.log)

	def enableTTS(self, enable):
		url = self.discord+"users/@me/settings"
		body = {"enable_tts_command": enable}
		return Wrapper.sendRequest(self.s, 'patch', url, body, log=self.log)

	'''
	Billing Settings - Billing
	'''
	def getBillingHistory(self, limit=20):
		url = self.discord+"users/@me/billing/payments?limit="+str(limit)
		return Wrapper.sendRequest(self.s, 'get', url, log=self.log)

	def getPaymentSources(self):
		url = self.discord+"users/@me/billing/payment-sources"
		return Wrapper.sendRequest(self.s, 'get', url, log=self.log)

	def getBillingSubscriptions(self):
		url = self.discord+"users/@me/billing/subscriptions"
		return Wrapper.sendRequest(self.s, 'get', url, log=self.log)

	def getStripeClientSecret(self): #for adding new payment methods. Stripe api wraps are not included because discum is just a discord api wrapper.
		url = self.discord+"users/@me/billing/stripe/setup-intents"
		return Wrapper.sendRequest(self.s, 'post', url, log=self.log)
	'''
	Billing Settings - Gift Inventory
	'''
	
	'''
	Logout
	'''
	def logout(self, provider=None, voip_provider=None):
		url = self.discord+"auth/logout"
		body = {"provider": provider, "voip_provider": voip_provider}
		return Wrapper.sendRequest(self.s, 'post', url, body, log=self.log)




#points to commands that help request info/actions using the gateway

class GuildRequest(object):
	def __init__(self, gatewayobject):
		self.gatewayobject = gatewayobject

	def lazyGuild(self, guild_id, channel_ranges, typing=None, threads=None, activities=None, members=None): #https://arandomnewaccount.gitlab.io/discord-unofficial-docs/lazy_guilds.html
		data = {
		    "op": self.gatewayobject.OPCODE.LAZY_REQUEST,
		    "d": {
		        "guild_id": guild_id,
		        "typing": typing,
		        "threads": threads,
		        "activities": activities,
		        "members": members,
		        "channels": channel_ranges,
		    },
		}
		if typing == None:
			data["d"].pop("typing")
		if threads == None:
			data["d"].pop("threads")
		if activities == None:
			data["d"].pop("activities")
		if members == None:
			data["d"].pop("members")
		self.gatewayobject.send(data)

	def searchGuildMembers(self, guild_ids, query, limit=10, presences=True):
		if isinstance(guild_ids, str):
			guild_ids = [guild_ids]
		self.gatewayobject.send({"op":self.gatewayobject.OPCODE.REQUEST_GUILD_MEMBERS,"d":{"guild_id":guildIDs,"query":query,"limit":limit,"presences":presences}})


import random
import time, datetime
import struct
import base64

class Client_UUID(object): #Huge thanks to github user fweak for helping me figure out the mystery of the client_uuid. made some discord "science" notes here: https://docs.google.com/document/d/1b5aDx7S1iLHoeb6B56izZakbXItA84gUjFzK-0OBwy0
    def __init__(self, userID, creationTime="now", eventNum=0):
        self.userID = int(userID)
        num = int(4294967296 * random.random())
        self.randomPrefix = num if num<=2147483647 else num-4294967296
        self.creationTime = int(time.mktime(datetime.datetime.now().timetuple()) * 1000) if creationTime == "now" else creationTime
        self.eventNum = eventNum
        self.UUID = ""

    def calculate(self, eventNum="default", userID="default", increment=True):
        if eventNum == "default":
            eventNum = self.eventNum
        if userID == "default":
            userID = self.userID
        else:
            userID = int(userID)

        buf = bytearray(struct.pack('24x'))
        buf[0:4] = struct.pack("<i", userID%4294967296 if userID%4294967296<=2147483647 else userID%4294967296-2147483647)
        buf[4:8] = struct.pack("<i", userID>>32)
        buf[8:12] = struct.pack("<i", self.randomPrefix)
        buf[12:16] = struct.pack("<i", self.creationTime%4294967296 if self.creationTime%4294967296<=2147483647 else self.creationTime%4294967296-2147483647)
        buf[16:20] = struct.pack("<i", self.creationTime>>32)
        buf[20:24] = struct.pack("<i", eventNum)

        if increment:
            self.eventNum += 1
        self.UUID = base64.b64encode(buf).decode('ascii')
        return self.UUID

    def refresh(self, resetEventNum=True):
        self.randomPrefix = num if num<=2147483647 else num-4294967296
        self.creationTime = int(time.mktime(datetime.datetime.now().timetuple()) * 1000) if creationTime == "now" else creationTime
        if resetEventNum:
            self.eventNum = 0
        return self.calculate()

    @staticmethod
    def parse(client_uuid):
        decoded_client_uuid = base64.b64decode(client_uuid)
        unpacked = []
        for i in range(6):
            unpacked.append(struct.unpack('<i', decoded_client_uuid[4*i:4*i+4])[0])
        UUIDdata = {}
        UUIDdata['userID'] = str((unpacked[1]<<32) + unpacked[0])
        UUIDdata['randomPrefix'] = unpacked[2]
        UUIDdata['creationTime'] = (unpacked[4]<<32) + unpacked[3]
        UUIDdata['eventNum'] = unpacked[5]
        return UUIDdata
